from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Mapping,
    NamedTuple,
    NewType,
    Optional,
    Sequence,
)

from dagster import _check as check
from dagster._core.definitions.asset_subset import AssetSubset, ValidAssetSubset
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.multi_dimensional_partitions import (
    MultiPartitionKey,
    MultiPartitionsDefinition,
    PartitionDimensionDefinition,
)
from dagster._core.definitions.partition import (
    AllPartitionsSubset,
    DefaultPartitionsSubset,
)
from dagster._core.definitions.time_window_partitions import (
    BaseTimeWindowPartitionsSubset,
    TimeWindow,
    TimeWindowPartitionsDefinition,
    get_time_partitions_def,
)
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.definitions.base_asset_graph import BaseAssetGraph, BaseAssetNode
    from dagster._core.definitions.data_version import CachingStaleStatusResolver
    from dagster._core.definitions.definitions_class import Definitions
    from dagster._core.definitions.partition import PartitionsDefinition
    from dagster._core.instance import DagsterInstance
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer


class TemporalContext(NamedTuple):
    """TemporalContext represents an effective time, used for business logic, and last_event_id
    which is used to identify that state of the event log at some point in time. Put another way,
    the value of a TemporalContext represents a point in time and a snapshot of the event log.

    Effective time: This is the effective time of the computation in terms of business logic,
    and it impacts the behavior of partitioning and partition mapping. For example,
    the "last" partition window of a given partitions definition, it is with
    respect to the effective time.

    Last event id: Our event log has a monotonically increasing event id. This is used to
    cursor the event log. This event_id is also propogated to derived tables to indicate
    when that record is valid.  This allows us to query the state of the event log
    at a given point in time.

    Note that insertion time of the last_event_id is not the same as the effective time.

    A last_event_id of None indicates that the reads will be volatile will immediately
    reflect any subsequent writes.
    """

    effective_dt: datetime
    last_event_id: Optional[int]


# We reserve the right to constraints on the AssetSubset that we are going to use
# in AssetSlice internals. Adding a NewType enforces that we do that conversion
# in one spot (see _slice_from_subset)
_AssetSliceCompatibleSubset = NewType("_AssetSliceCompatibleSubset", ValidAssetSubset)


def _slice_from_subset(asset_graph_view: "AssetGraphView", subset: AssetSubset) -> "AssetSlice":
    valid_subset = subset.as_valid(
        asset_graph_view.asset_graph.get(subset.asset_key).partitions_def
    )
    return AssetSlice(asset_graph_view, _AssetSliceCompatibleSubset(valid_subset))


class AssetSlice:
    """An asset slice represents a set of partitions for a given asset key. It is
    tied to a particular instance of an AssetGraphView, and is read-only.

    With an asset slice you are able to traverse the set of partitions resident
    in an asset graph at any point in time.

    ```python
        asset_graph_view_t0 = AssetGraphView.for_test(defs, effective_dt=some_date())

        some_asset_slice = asset_graph_view_to.get_asset_slice(some_asset.key)

        for parent_slice in some_asset_slice.compute_parent_slices().values():
            # do something with the parent slice
    ```

    AssetSlice is read-only and tied to a specific AssetGraphView. Therefore
    we can safely use cached methods and properties at will. However different methods
    have different performance characterics so we have the following conventions:

    Naming conventions
    * Properties are guaranteed to be fast.
    * Methods prefixed with `get_` do some work in-memory but are not hugely expensive.
    * Methods prefixed with `compute_` do potentially expensive work, like compute
      partition mappings and query the instance.

      We also use this prefix to indicate that they fully materialize partition sets
      These can potentially be very expensive if the underlying partition set has
      an in-memory representation that involves large time windows. I.e. if the
      underlying PartitionsSubset in the ValidAssetSubset is a
      TimeWindowPartitionsSubset. Usage of these methods should be avoided if
      possible if you are potentially dealing with slices with large time-based
      partition windows.
    """

    def __init__(
        self, asset_graph_view: "AssetGraphView", compatible_subset: _AssetSliceCompatibleSubset
    ):
        self._asset_graph_view = asset_graph_view
        self._compatible_subset = compatible_subset

    def convert_to_valid_asset_subset(self) -> ValidAssetSubset:
        return self._compatible_subset

    # only works for partitioned assets for now
    def compute_partition_keys(self) -> AbstractSet[str]:
        return {
            check.not_none(akpk.partition_key, "No None partition keys")
            for akpk in self._compatible_subset.asset_partitions
        }

    def expensively_compute_asset_partitions(self) -> AbstractSet[AssetKeyPartitionKey]:
        # this method requires computing all partition keys of the definition, which
        # may be expensive
        return self._compatible_subset.asset_partitions

    @property
    def asset_key(self) -> AssetKey:
        return self._compatible_subset.asset_key

    @property
    def parent_keys(self) -> AbstractSet[AssetKey]:
        return self._asset_graph_view.asset_graph.get(self.asset_key).parent_keys

    @property
    def child_keys(self) -> AbstractSet[AssetKey]:
        return self._asset_graph_view.asset_graph.get(self.asset_key).child_keys

    @property
    def _partitions_def(self) -> Optional["PartitionsDefinition"]:
        return self._asset_graph_view.asset_graph.get(self.asset_key).partitions_def

    @cached_method
    def compute_parent_slice(self, parent_asset_key: AssetKey) -> "AssetSlice":
        return self._asset_graph_view.compute_parent_asset_slice(parent_asset_key, self)

    @cached_method
    def compute_child_slice(self, child_asset_key: AssetKey) -> "AssetSlice":
        return self._asset_graph_view.compute_child_asset_slice(child_asset_key, self)

    @cached_method
    def compute_parent_slices(self) -> Mapping[AssetKey, "AssetSlice"]:
        return {ak: self.compute_parent_slice(ak) for ak in self.parent_keys}

    @cached_method
    def compute_child_slices(self) -> Mapping[AssetKey, "AssetSlice"]:
        return {ak: self.compute_child_slice(ak) for ak in self.child_keys}

    def compute_difference(self, other: "AssetSlice") -> "AssetSlice":
        return _slice_from_subset(
            self._asset_graph_view, self._compatible_subset - other.convert_to_valid_asset_subset()
        )

    def compute_union(self, other: "AssetSlice") -> "AssetSlice":
        return _slice_from_subset(
            self._asset_graph_view, self._compatible_subset | other.convert_to_valid_asset_subset()
        )

    def compute_intersection(self, other: "AssetSlice") -> "AssetSlice":
        return _slice_from_subset(
            self._asset_graph_view, self._compatible_subset & other.convert_to_valid_asset_subset()
        )

    def compute_intersection_with_partition_keys(
        self, partition_keys: AbstractSet[str]
    ) -> "AssetSlice":
        """Return a new AssetSlice with only the given partition keys if they are in the slice."""
        return self._asset_graph_view.compute_intersection_with_partition_keys(partition_keys, self)

    @property
    def time_windows(self) -> Sequence[TimeWindow]:
        """Get the time windows for the asset slice. Only supports explicitly time-windowed partitions for now."""
        tw_partitions_def = check.not_none(
            self._time_window_partitions_def_in_context(), "Must be time windowed."
        )

        if isinstance(self._compatible_subset.subset_value, BaseTimeWindowPartitionsSubset):
            return self._compatible_subset.subset_value.included_time_windows
        elif isinstance(self._compatible_subset.subset_value, AllPartitionsSubset):
            last_tw = tw_partitions_def.get_last_partition_window(
                self._asset_graph_view.effective_dt
            )
            return [TimeWindow(datetime.min, last_tw.end)] if last_tw else []
        elif isinstance(self._compatible_subset.subset_value, DefaultPartitionsSubset):
            check.inst(
                self._partitions_def,
                MultiPartitionsDefinition,
                "Must be multi-partition if we got here.",
            )
            tw_partition_keys = set()
            for multi_partition_key in check.is_list(
                list(self._compatible_subset.subset_value.get_partition_keys()),
                MultiPartitionKey,
                "Keys must be multi partition keys.",
            ):
                tm_partition_key = next(iter(multi_partition_key.keys_by_dimension.values()))
                tw_partition_keys.add(tm_partition_key)

            subset_from_tw = tw_partitions_def.subset_with_partition_keys(tw_partition_keys)
            check.inst(
                subset_from_tw, BaseTimeWindowPartitionsSubset, "Must be time window subset."
            )
            if isinstance(subset_from_tw, BaseTimeWindowPartitionsSubset):
                return subset_from_tw.included_time_windows
            else:
                check.failed(
                    f"Unsupported subset value in generated subset {self._compatible_subset.subset_value} created by keys {tw_partition_keys}"
                )

        check.failed(f"Unsupported subset value: {self._compatible_subset.subset_value}")

    def _time_window_partitions_def_in_context(self) -> Optional[TimeWindowPartitionsDefinition]:
        pd = self._partitions_def
        if isinstance(pd, TimeWindowPartitionsDefinition):
            return pd
        if isinstance(pd, MultiPartitionsDefinition):
            return pd.time_window_partitions_def if pd.has_time_window_dimension else None
        return None

    @property
    def is_empty(self) -> bool:
        return self._compatible_subset.is_empty


class AssetGraphView:
    """The Asset Graph View. It is a view of the asset graph from the perspective of a specific
    temporal context.

    If the user wants to get a new view of the asset graph with a new effective date or last event
    id, they should create a new instance of an AssetGraphView. If they do not they will get
    incorrect results because the AssetGraphView and its associated classes (like AssetSlice)
    cache results based on the effective date and last event id.

    ```python
        # in a test case
        asset_graph_view_t0 = AssetGraphView.for_test(defs, effective_dt=some_date())

        #
        # call materialize on an asset in defs
        #
        # must create a new AssetGraphView to get the correct results,
        # asset_graph_view_t1 will not reflect the new materialization
        asset_graph_view_t1 = AssetGraphView.for_test(defs, effective_dt=some_date())
    ```

    """

    @staticmethod
    def for_test(
        defs: "Definitions",
        instance: Optional["DagsterInstance"] = None,
        effective_dt: Optional[datetime] = None,
        last_event_id: Optional[int] = None,
    ):
        import pendulum

        from dagster._core.definitions.data_version import CachingStaleStatusResolver
        from dagster._core.instance import DagsterInstance

        stale_resolver = CachingStaleStatusResolver(
            instance=instance or DagsterInstance.ephemeral(),
            asset_graph=defs.get_asset_graph(),
        )
        check.invariant(stale_resolver.instance_queryer, "Ensure instance queryer is constructed")
        return AssetGraphView(
            stale_resolver=stale_resolver,
            temporal_context=TemporalContext(
                effective_dt=effective_dt or pendulum.now(),
                last_event_id=last_event_id,
            ),
        )

    def __init__(
        self,
        *,
        temporal_context: TemporalContext,
        stale_resolver: "CachingStaleStatusResolver",
    ):
        # Current these properties have the ability to be lazily constructed.
        # We instead are going to try to retain the invariant that they are
        # constructed upfront so that initialization time is well-understood
        # and deterministic. If there are cheap operations that do not
        # require these instances, we should move them to a different abstraction.

        # ensure it is already constructed rather than created on demand
        check.invariant(stale_resolver._instance_queryer)  # noqa: SLF001
        # ensure it is already constructed rather than created on demand
        check.invariant(stale_resolver._asset_graph)  # noqa: SLF001

        # stale resolver has a CachingInstanceQueryer which has a DagsterInstance
        # so just passing the CachingStaleStatusResolver is enough
        self._stale_resolver = stale_resolver
        self._temporal_context = temporal_context

    @property
    def effective_dt(self) -> datetime:
        return self._temporal_context.effective_dt

    @property
    def last_event_id(self) -> Optional[int]:
        return self._temporal_context.last_event_id

    @property
    def asset_graph(self) -> "BaseAssetGraph[BaseAssetNode]":
        return self._stale_resolver.asset_graph

    @property
    def _queryer(self) -> "CachingInstanceQueryer":
        return self._stale_resolver.instance_queryer

    def _get_partitions_def(self, asset_key: "AssetKey") -> Optional["PartitionsDefinition"]:
        return self.asset_graph.get(asset_key).partitions_def

    def get_asset_slice(self, asset_key: "AssetKey") -> "AssetSlice":
        # not compute_asset_slice because dynamic partitions store
        # is just passed to AssetSubset.all, not invoked
        return _slice_from_subset(
            self,
            AssetSubset.all(
                asset_key=asset_key,
                partitions_def=self._get_partitions_def(asset_key),
                dynamic_partitions_store=self._queryer,
                current_time=self.effective_dt,
            ),
        )

    def get_asset_slice_from_subset(self, subset: AssetSubset) -> Optional["AssetSlice"]:
        if subset.is_compatible_with_partitions_def(self._get_partitions_def(subset.asset_key)):
            return _slice_from_subset(self, subset)
        else:
            return None

    def get_asset_slice_from_valid_subset(self, subset: ValidAssetSubset) -> "AssetSlice":
        return _slice_from_subset(self, subset)

    def get_asset_slice_from_asset_partitions(
        self, asset_partitions: AbstractSet[AssetKeyPartitionKey]
    ) -> "AssetSlice":
        asset_keys = {akpk.asset_key for akpk in asset_partitions}
        check.invariant(len(asset_keys) == 1, "Must have exactly one asset key")
        asset_key = asset_keys.pop()
        return _slice_from_subset(
            self,
            ValidAssetSubset.from_asset_partitions_set(
                asset_key=asset_key,
                partitions_def=self._get_partitions_def(asset_key),
                asset_partitions_set=asset_partitions,
            ),
        )

    def create_empty_slice(self, asset_key: AssetKey) -> AssetSlice:
        return _slice_from_subset(
            self,
            AssetSubset.empty(asset_key, self._get_partitions_def(asset_key)),
        )

    def compute_parent_asset_slice(
        self, parent_asset_key: AssetKey, asset_slice: AssetSlice
    ) -> AssetSlice:
        return _slice_from_subset(
            self,
            self.asset_graph.get_parent_asset_subset(
                dynamic_partitions_store=self._queryer,
                parent_asset_key=parent_asset_key,
                child_asset_subset=asset_slice.convert_to_valid_asset_subset(),
                current_time=self.effective_dt,
            ),
        )

    def compute_child_asset_slice(
        self, child_asset_key: "AssetKey", asset_slice: AssetSlice
    ) -> "AssetSlice":
        return _slice_from_subset(
            self,
            self.asset_graph.get_child_asset_subset(
                dynamic_partitions_store=self._queryer,
                child_asset_key=child_asset_key,
                current_time=self.effective_dt,
                parent_asset_subset=asset_slice.convert_to_valid_asset_subset(),
            ),
        )

    def compute_intersection_with_partition_keys(
        self, partition_keys: AbstractSet[str], asset_slice: AssetSlice
    ) -> "AssetSlice":
        """Return a new AssetSlice with only the given partition keys if they are in the slice."""
        partitions_def = check.not_none(
            self._get_partitions_def(asset_slice.asset_key), "Must have partitions def"
        )
        for partition_key in partition_keys:
            if not partitions_def.has_partition_key(
                partition_key,
                current_time=self.effective_dt,
                dynamic_partitions_store=self._queryer,
            ):
                check.failed(
                    f"Partition key {partition_key} not in partitions def {partitions_def}"
                )

        return _slice_from_subset(
            self,
            asset_slice.convert_to_valid_asset_subset()
            & AssetSubset.from_partition_keys(
                asset_slice.asset_key, partitions_def, partition_keys
            ),
        )

    def compute_latest_time_window_slice(
        self, asset_key: AssetKey, lookback_delta: Optional[timedelta] = None
    ) -> AssetSlice:
        """Compute the slice of the asset which exists within the latest time partition window. If
        the asset has no time dimension, this will always return the full slice. If
        lookback_delta is provided, all partitions that are up to that timedelta before the
        end of the latest time partition window will be included.
        """
        partitions_def = self._get_partitions_def(asset_key)
        time_partitions_def = get_time_partitions_def(partitions_def)
        if time_partitions_def is None:
            # if the asset has no time dimension, then return a full slice
            return self.get_asset_slice(asset_key)

        latest_time_window = time_partitions_def.get_last_partition_window(self.effective_dt)
        if latest_time_window is None:
            return self.create_empty_slice(asset_key)

        # the time window in which to look for partitions
        time_window = (
            TimeWindow(
                start=max(
                    # do not look before the start of the definition
                    time_partitions_def.start,
                    latest_time_window.end - lookback_delta,
                ),
                end=latest_time_window.end,
            )
            if lookback_delta
            else latest_time_window
        )

        if isinstance(partitions_def, TimeWindowPartitionsDefinition):
            return self._build_time_partition_slice(asset_key, partitions_def, time_window)
        elif isinstance(partitions_def, MultiPartitionsDefinition):
            return self._build_multi_partition_slice(
                asset_key, self._get_multi_dim_info(asset_key), time_window
            )
        else:
            check.failed(f"Unsupported partitions_def: {partitions_def}")

    def compute_missing_subslice(
        self, asset_key: "AssetKey", from_slice: "AssetSlice"
    ) -> "AssetSlice":
        """Returns a slice which is the subset of the input slice that has never been materialized
        (if it is a materializable asset) or observered (if it is an observable asset).
        """
        # TODO: this logic should be simplified once we have a unified way of detecting both
        # materializations and observations through the parittion status cache. at that point, the
        # definition will slightly change to search for materializations and observations regardless
        # of the materializability of the asset
        if self.asset_graph.get(asset_key).is_materializable:
            # cheap call which takes advantage of the partition status cache
            materialized_subset = self._queryer.get_materialized_asset_subset(asset_key=asset_key)
            materialized_slice = self.get_asset_slice_from_valid_subset(materialized_subset)
            return from_slice.compute_difference(materialized_slice)
        else:
            # more expensive call
            missing_asset_partitions = {
                ap
                for ap in from_slice.convert_to_valid_asset_subset().asset_partitions
                if not self._queryer.asset_partition_has_materialization_or_observation(ap)
            }
            missing_subset = ValidAssetSubset.from_asset_partitions_set(
                asset_key, self._get_partitions_def(asset_key), missing_asset_partitions
            )
            return self.get_asset_slice_from_valid_subset(missing_subset)

    @cached_method
    def compute_in_progress_asset_slice(self, *, asset_key: "AssetKey") -> "AssetSlice":
        return _slice_from_subset(
            self, self._queryer.get_in_progress_asset_subset(asset_key=asset_key)
        )

    @cached_method
    def compute_failed_asset_slice(self, *, asset_key: "AssetKey") -> "AssetSlice":
        return _slice_from_subset(self, self._queryer.get_failed_asset_subset(asset_key=asset_key))

    @cached_method
    def compute_updated_since_cursor_slice(
        self, *, asset_key: AssetKey, cursor: Optional[int]
    ) -> AssetSlice:
        subset = self._queryer.get_asset_subset_updated_after_cursor(
            asset_key=asset_key, after_cursor=cursor
        )
        return self.get_asset_slice_from_valid_subset(subset)

    @cached_method
    def compute_parent_updated_since_cursor_slice(
        self, *, asset_key: AssetKey, cursor: Optional[int]
    ) -> AssetSlice:
        result_slice = self.create_empty_slice(asset_key)
        for parent_key in self.asset_graph.get(asset_key).parent_keys:
            result_slice = result_slice.compute_union(
                self.compute_updated_since_cursor_slice(
                    asset_key=parent_key, cursor=cursor
                ).compute_child_slice(asset_key)
            )
        return result_slice

    class MultiDimInfo(NamedTuple):
        tw_dim: PartitionDimensionDefinition
        secondary_dim: PartitionDimensionDefinition

        @property
        def tw_partition_def(self) -> TimeWindowPartitionsDefinition:
            return check.inst(
                self.tw_dim.partitions_def,
                TimeWindowPartitionsDefinition,
            )

        @property
        def secondary_partition_def(self) -> "PartitionsDefinition":
            return self.secondary_dim.partitions_def

    def _get_multi_dim_info(self, asset_key: AssetKey) -> "MultiDimInfo":
        partitions_def = check.inst(
            self._get_partitions_def(asset_key),
            MultiPartitionsDefinition,
        )
        return self.MultiDimInfo(
            tw_dim=partitions_def.time_window_dimension,
            secondary_dim=partitions_def.secondary_dimension,
        )

    def _build_time_partition_slice(
        self,
        asset_key: AssetKey,
        partitions_def: TimeWindowPartitionsDefinition,
        time_window: TimeWindow,
    ) -> "AssetSlice":
        return self.get_asset_slice(asset_key).compute_intersection_with_partition_keys(
            set(partitions_def.get_partition_keys_in_time_window(time_window))
        )

    def _build_multi_partition_slice(
        self, asset_key: AssetKey, multi_dim_info: MultiDimInfo, time_window: TimeWindow
    ) -> "AssetSlice":
        # Note: Potential perf improvement here. There is no way to encode a cartesian product
        # in the underlying PartitionsSet. We could add a specialized PartitionsSubset
        # subclass that itself composed two PartitionsSubset to avoid materializing the entire
        # partitions range.
        return self.get_asset_slice(asset_key).compute_intersection_with_partition_keys(
            {
                MultiPartitionKey(
                    {
                        multi_dim_info.tw_dim.name: tw_pk,
                        multi_dim_info.secondary_dim.name: secondary_pk,
                    }
                )
                for tw_pk in multi_dim_info.tw_partition_def.get_partition_keys_in_time_window(
                    time_window
                )
                for secondary_pk in multi_dim_info.secondary_partition_def.get_partition_keys(
                    current_time=self.effective_dt,
                    dynamic_partitions_store=self._queryer,
                )
            }
        )



import logging
import sys
from typing import Iterable, Mapping, Optional, Sequence, cast

from dagster._core.execution.asset_backfill import execute_asset_backfill_iteration
from dagster._core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster._core.execution.job_backfill import execute_job_backfill_iteration
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.utils import DaemonErrorCapture
from dagster._utils.error import SerializableErrorInfo


def execute_backfill_iteration(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    debug_crash_flags: Optional[Mapping[str, int]] = None,
) -> Iterable[Optional[SerializableErrorInfo]]:
    instance = workspace_process_context.instance

    in_progress_backfills = instance.get_backfills(status=BulkActionStatus.REQUESTED)
    canceling_backfills = instance.get_backfills(status=BulkActionStatus.CANCELING)

    if not in_progress_backfills and not canceling_backfills:
        logger.debug("No backfill jobs in progress or canceling.")
        yield None
        return

    backfill_jobs = [*in_progress_backfills, *canceling_backfills]

    yield from execute_backfill_jobs(
        workspace_process_context, logger, backfill_jobs, debug_crash_flags
    )


def execute_backfill_jobs(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    backfill_jobs: Sequence[PartitionBackfill],
    debug_crash_flags: Optional[Mapping[str, int]] = None,
) -> Iterable[Optional[SerializableErrorInfo]]:
    instance = workspace_process_context.instance

    for backfill_job in backfill_jobs:
        backfill_id = backfill_job.backfill_id

        # refetch, in case the backfill was updated in the meantime
        backfill = cast(PartitionBackfill, instance.get_backfill(backfill_id))
        # create a logger that will always include the backfill_id as an `extra`

        backfill_logger = cast(
            logging.Logger,
            logging.LoggerAdapter(logger, extra={"backfill_id": backfill.backfill_id}),
        )

        try:
            if backfill.is_asset_backfill:
                yield from execute_asset_backfill_iteration(
                    backfill, backfill_logger, workspace_process_context, instance
                )
            else:
                yield from execute_job_backfill_iteration(
                    backfill,
                    backfill_logger,
                    workspace_process_context,
                    debug_crash_flags,
                    instance,
                )
        except Exception:
            error_info = DaemonErrorCapture.on_exception(
                sys.exc_info(),
                logger=backfill_logger,
                log_message=f"Backfill failed for {backfill.backfill_id}",
            )
            instance.update_backfill(
                backfill.with_status(BulkActionStatus.FAILED).with_error(error_info)
            )
            yield error_info



import operator
from collections import defaultdict
from datetime import datetime
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Union,
    cast,
)

from dagster import _check as check
from dagster._core.definitions.partition import (
    PartitionsDefinition,
    PartitionsSubset,
)
from dagster._core.errors import (
    DagsterDefinitionChangedDeserializationError,
)
from dagster._core.instance import DynamicPartitionsStore
from dagster._serdes.serdes import (
    NamedTupleSerializer,
    SerializableNonScalarKeyMapping,
    whitelist_for_serdes,
)

from .asset_subset import AssetSubset
from .base_asset_graph import BaseAssetGraph
from .events import AssetKey, AssetKeyPartitionKey


class PartitionsSubsetMappingNamedTupleSerializer(NamedTupleSerializer):
    """Serializes NamedTuples with fields that are mappings containing PartitionsSubsets.

    This is necessary because PartitionKeysTimeWindowPartitionsSubsets are not serializable,
    so we convert them to TimeWindowPartitionsSubsets.
    """

    def before_pack(self, value: NamedTuple) -> NamedTuple:
        replaced_value_by_field_name = {}
        for field_name, field_value in value._asdict().items():
            if isinstance(field_value, Mapping) and all(
                isinstance(v, PartitionsSubset) for v in field_value.values()
            ):
                # PartitionKeysTimeWindowPartitionsSubsets are not serializable, so
                # we convert them to TimeWindowPartitionsSubsets
                subsets_by_key = {k: v.to_serializable_subset() for k, v in field_value.items()}

                # If the mapping is keyed by AssetKey wrap it in a SerializableNonScalarKeyMapping
                # so it can be serialized. This can be expanded to other key types in the future.
                if all(isinstance(k, AssetKey) for k in subsets_by_key.keys()):
                    replaced_value_by_field_name[field_name] = SerializableNonScalarKeyMapping(
                        subsets_by_key
                    )

        return value._replace(**replaced_value_by_field_name)


@whitelist_for_serdes(serializer=PartitionsSubsetMappingNamedTupleSerializer)
class AssetGraphSubset(NamedTuple):
    partitions_subsets_by_asset_key: Mapping[AssetKey, PartitionsSubset] = {}
    non_partitioned_asset_keys: AbstractSet[AssetKey] = set()

    @property
    def asset_keys(self) -> AbstractSet[AssetKey]:
        return {
            key for key, subset in self.partitions_subsets_by_asset_key.items() if len(subset) > 0
        } | self.non_partitioned_asset_keys

    @property
    def num_partitions_and_non_partitioned_assets(self) -> int:
        return len(self.non_partitioned_asset_keys) + sum(
            len(subset) for subset in self.partitions_subsets_by_asset_key.values()
        )

    def get_asset_subset(self, asset_key: AssetKey, asset_graph: BaseAssetGraph) -> AssetSubset:
        """Returns an AssetSubset representing the subset of a specific asset that this
        AssetGraphSubset contains.
        """
        partitions_def = asset_graph.get(asset_key).partitions_def
        if partitions_def is None:
            return AssetSubset(
                asset_key=asset_key, value=asset_key in self.non_partitioned_asset_keys
            )
        else:
            return AssetSubset(
                asset_key=asset_key,
                value=self.partitions_subsets_by_asset_key.get(
                    asset_key, partitions_def.empty_subset()
                ),
            )

    def get_partitions_subset(
        self, asset_key: AssetKey, asset_graph: Optional[BaseAssetGraph] = None
    ) -> PartitionsSubset:
        if asset_graph:
            partitions_def = asset_graph.get(asset_key).partitions_def
            if partitions_def is None:
                check.failed("Can only call get_partitions_subset on a partitioned asset")

            return self.partitions_subsets_by_asset_key.get(
                asset_key, partitions_def.empty_subset()
            )
        else:
            return self.partitions_subsets_by_asset_key[asset_key]

    def iterate_asset_partitions(self) -> Iterable[AssetKeyPartitionKey]:
        for (
            asset_key,
            partitions_subset,
        ) in self.partitions_subsets_by_asset_key.items():
            for partition_key in partitions_subset.get_partition_keys():
                yield AssetKeyPartitionKey(asset_key, partition_key)

        for asset_key in self.non_partitioned_asset_keys:
            yield AssetKeyPartitionKey(asset_key, None)

    def __contains__(self, asset: Union[AssetKey, AssetKeyPartitionKey]) -> bool:
        """If asset is an AssetKeyPartitionKey, check if the given AssetKeyPartitionKey is in the
        subset. If asset is an AssetKey, check if any of partitions of the given AssetKey are in
        the subset.
        """
        if isinstance(asset, AssetKey):
            # check if any keys are in the subset
            partitions_subset = self.partitions_subsets_by_asset_key.get(asset)
            return (partitions_subset is not None and len(partitions_subset) > 0) or (
                asset in self.non_partitioned_asset_keys
            )
        elif asset.partition_key is None:
            return asset.asset_key in self.non_partitioned_asset_keys
        else:
            partitions_subset = self.partitions_subsets_by_asset_key.get(asset.asset_key)
            return partitions_subset is not None and asset.partition_key in partitions_subset

    def to_storage_dict(
        self, dynamic_partitions_store: DynamicPartitionsStore, asset_graph: BaseAssetGraph
    ) -> Mapping[str, object]:
        return {
            "partitions_subsets_by_asset_key": {
                key.to_user_string(): value.serialize()
                for key, value in self.partitions_subsets_by_asset_key.items()
            },
            "serializable_partitions_def_ids_by_asset_key": {
                key.to_user_string(): check.not_none(
                    asset_graph.get(key).partitions_def
                ).get_serializable_unique_identifier(
                    dynamic_partitions_store=dynamic_partitions_store
                )
                for key, _ in self.partitions_subsets_by_asset_key.items()
            },
            "partitions_def_class_names_by_asset_key": {
                key.to_user_string(): check.not_none(
                    asset_graph.get(key).partitions_def
                ).__class__.__name__
                for key, _ in self.partitions_subsets_by_asset_key.items()
            },
            "non_partitioned_asset_keys": [
                key.to_user_string() for key in self.non_partitioned_asset_keys
            ],
        }

    def _oper(self, other: "AssetGraphSubset", oper: Callable[..., Any]) -> "AssetGraphSubset":
        """Returns the AssetGraphSubset that results from applying the given operator to the set of
        asset partitions in self and other.

        Note: Not all operators are supported on the underlying PartitionsSubset objects.
        """
        result_partition_subsets_by_asset_key = {**self.partitions_subsets_by_asset_key}
        result_non_partitioned_asset_keys = set(self.non_partitioned_asset_keys)

        for asset_key in other.asset_keys:
            if asset_key in other.non_partitioned_asset_keys:
                check.invariant(asset_key not in self.partitions_subsets_by_asset_key)
                result_non_partitioned_asset_keys = oper(
                    result_non_partitioned_asset_keys, {asset_key}
                )
            else:
                check.invariant(asset_key not in self.non_partitioned_asset_keys)
                subset = (
                    self.get_partitions_subset(asset_key)
                    if asset_key in self.partitions_subsets_by_asset_key
                    else None
                )

                other_subset = other.get_partitions_subset(asset_key)

                if other_subset is not None and subset is not None:
                    result_partition_subsets_by_asset_key[asset_key] = oper(subset, other_subset)

                # Special case operations if either subset is None
                if subset is None and other_subset is not None and oper == operator.or_:
                    result_partition_subsets_by_asset_key[asset_key] = other_subset
                elif subset is not None and other_subset is None and oper == operator.and_:
                    del result_partition_subsets_by_asset_key[asset_key]

        return AssetGraphSubset(
            partitions_subsets_by_asset_key=result_partition_subsets_by_asset_key,
            non_partitioned_asset_keys=result_non_partitioned_asset_keys,
        )

    def __or__(self, other: "AssetGraphSubset") -> "AssetGraphSubset":
        return self._oper(other, operator.or_)

    def __sub__(self, other: "AssetGraphSubset") -> "AssetGraphSubset":
        return self._oper(other, operator.sub)

    def __and__(self, other: "AssetGraphSubset") -> "AssetGraphSubset":
        return self._oper(other, operator.and_)

    def filter_asset_keys(self, asset_keys: AbstractSet[AssetKey]) -> "AssetGraphSubset":
        return AssetGraphSubset(
            partitions_subsets_by_asset_key={
                asset_key: subset
                for asset_key, subset in self.partitions_subsets_by_asset_key.items()
                if asset_key in asset_keys
            },
            non_partitioned_asset_keys=self.non_partitioned_asset_keys & asset_keys,
        )

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, AssetGraphSubset)
            and self.partitions_subsets_by_asset_key == other.partitions_subsets_by_asset_key
            and self.non_partitioned_asset_keys == other.non_partitioned_asset_keys
        )

    def __repr__(self) -> str:
        return (
            "AssetGraphSubset("
            f"non_partitioned_asset_keys={self.non_partitioned_asset_keys}, "
            f"partitions_subsets_by_asset_key={self.partitions_subsets_by_asset_key}"
            ")"
        )

    @classmethod
    def from_asset_partition_set(
        cls,
        asset_partitions_set: AbstractSet[AssetKeyPartitionKey],
        asset_graph: BaseAssetGraph,
    ) -> "AssetGraphSubset":
        partitions_by_asset_key = defaultdict(set)
        non_partitioned_asset_keys = set()
        for asset_key, partition_key in asset_partitions_set:
            if partition_key is not None:
                partitions_by_asset_key[asset_key].add(partition_key)
            else:
                non_partitioned_asset_keys.add(asset_key)

        return AssetGraphSubset(
            partitions_subsets_by_asset_key={
                asset_key: (
                    cast(PartitionsDefinition, asset_graph.get(asset_key).partitions_def)
                    .empty_subset()
                    .with_partition_keys(partition_keys)
                )
                for asset_key, partition_keys in partitions_by_asset_key.items()
            },
            non_partitioned_asset_keys=non_partitioned_asset_keys,
        )

    @classmethod
    def can_deserialize(
        cls, serialized_dict: Mapping[str, Any], asset_graph: BaseAssetGraph
    ) -> bool:
        serializable_partitions_ids = serialized_dict.get(
            "serializable_partitions_def_ids_by_asset_key", {}
        )

        partitions_def_class_names_by_asset_key = serialized_dict.get(
            "partitions_def_class_names_by_asset_key", {}
        )

        for key, value in serialized_dict["partitions_subsets_by_asset_key"].items():
            asset_key = AssetKey.from_user_string(key)
            if not asset_graph.has(asset_key):
                # Asset had a definition at storage time, but no longer does
                return False

            partitions_def = asset_graph.get(asset_key).partitions_def
            if partitions_def is None:
                # Asset had a partitions definition at storage time, but no longer does
                return False

            if not partitions_def.can_deserialize_subset(
                value,
                serialized_partitions_def_unique_id=serializable_partitions_ids.get(key),
                serialized_partitions_def_class_name=partitions_def_class_names_by_asset_key.get(
                    key
                ),
            ):
                return False

        return True

    @classmethod
    def from_storage_dict(
        cls,
        serialized_dict: Mapping[str, Any],
        asset_graph: BaseAssetGraph,
        allow_partial: bool = False,
    ) -> "AssetGraphSubset":
        serializable_partitions_ids = serialized_dict.get(
            "serializable_partitions_def_ids_by_asset_key", {}
        )

        partitions_def_class_names_by_asset_key = serialized_dict.get(
            "partitions_def_class_names_by_asset_key", {}
        )
        partitions_subsets_by_asset_key: Dict[AssetKey, PartitionsSubset] = {}
        for key, value in serialized_dict["partitions_subsets_by_asset_key"].items():
            asset_key = AssetKey.from_user_string(key)

            if asset_key not in asset_graph.all_asset_keys:
                if not allow_partial:
                    raise DagsterDefinitionChangedDeserializationError(
                        f"Asset {key} existed at storage-time, but no longer does"
                    )
                continue

            partitions_def = asset_graph.get(asset_key).partitions_def

            if partitions_def is None:
                if not allow_partial:
                    raise DagsterDefinitionChangedDeserializationError(
                        f"Asset {key} had a PartitionsDefinition at storage-time, but no longer"
                        " does"
                    )
                continue

            if not partitions_def.can_deserialize_subset(
                value,
                serialized_partitions_def_unique_id=serializable_partitions_ids.get(key),
                serialized_partitions_def_class_name=partitions_def_class_names_by_asset_key.get(
                    key
                ),
            ):
                if not allow_partial:
                    raise DagsterDefinitionChangedDeserializationError(
                        f"Cannot deserialize stored partitions subset for asset {key}. This likely"
                        " indicates that the partitions definition has changed since this was"
                        " stored"
                    )
                continue

            partitions_subsets_by_asset_key[asset_key] = partitions_def.deserialize_subset(value)

        non_partitioned_asset_keys = {
            AssetKey.from_user_string(key) for key in serialized_dict["non_partitioned_asset_keys"]
        } & asset_graph.all_asset_keys

        return AssetGraphSubset(
            partitions_subsets_by_asset_key=partitions_subsets_by_asset_key,
            non_partitioned_asset_keys=non_partitioned_asset_keys,
        )

    @classmethod
    def all(
        cls,
        asset_graph: BaseAssetGraph,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
    ) -> "AssetGraphSubset":
        return cls.from_asset_keys(
            asset_graph.materializable_asset_keys,
            asset_graph,
            dynamic_partitions_store,
            current_time,
        )

    @classmethod
    def from_asset_keys(
        cls,
        asset_keys: Iterable[AssetKey],
        asset_graph: BaseAssetGraph,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
    ) -> "AssetGraphSubset":
        partitions_subsets_by_asset_key: Dict[AssetKey, PartitionsSubset] = {}
        non_partitioned_asset_keys: Set[AssetKey] = set()

        for asset_key in asset_keys:
            partitions_def = asset_graph.get(asset_key).partitions_def
            if partitions_def:
                partitions_subsets_by_asset_key[asset_key] = (
                    partitions_def.empty_subset().with_partition_keys(
                        partitions_def.get_partition_keys(
                            dynamic_partitions_store=dynamic_partitions_store,
                            current_time=current_time,
                        )
                    )
                )
            else:
                non_partitioned_asset_keys.add(asset_key)

        return AssetGraphSubset(
            partitions_subsets_by_asset_key=partitions_subsets_by_asset_key,
            non_partitioned_asset_keys=non_partitioned_asset_keys,
        )



from enum import Enum
from typing import TYPE_CHECKING, Any, Iterable, Mapping, NamedTuple, Optional, Sequence

import dagster._check as check
from dagster._annotations import PublicAttr, experimental_param
from dagster._serdes.serdes import whitelist_for_serdes

from .auto_materialize_policy import AutoMaterializePolicy
from .events import (
    AssetKey,
    CoercibleToAssetKey,
)
from .freshness_policy import FreshnessPolicy
from .utils import validate_tags_strict

if TYPE_CHECKING:
    from dagster._core.definitions.asset_dep import AssetDep, CoercibleToAssetDep

# SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE lives on the metadata of an asset
# (which currently ends up on the Output associated with the asset key)
# whih encodes the execution type the of asset. "Unexecutable" assets are assets
# that cannot be materialized in Dagster, but can have events in the event
# log keyed off of them, making Dagster usable as a observability and lineage tool
# for externally materialized assets.
SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE = "dagster/asset_execution_type"

# SYSTEM_METADATA_KEY_AUTO_OBSERVE_INTERVAL_MINUTES lives on the metadata of
# external assets resulting from a source asset conversion. It contains the
# `auto_observe_interval_minutes` value from the source asset and is consulted
# in the auto-materialize daemon. It should eventually be eliminated in favor
# of an implementation of `auto_observe_interval_minutes` in terms of
# `AutoMaterializeRule`.
SYSTEM_METADATA_KEY_AUTO_OBSERVE_INTERVAL_MINUTES = "dagster/auto_observe_interval_minutes"

# SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET lives on the metadata of external assets that are
# created for undefined but referenced assets during asset graph normalization. For example, in the
# below definitions, `foo` is referenced by upstream `bar` but has no corresponding definition:
#
#
#     @asset(deps=["foo"])
#     def bar(context: AssetExecutionContext):
#         ...
#
#     defs=Definitions(assets=[bar])
#
# During normalization we create a "stub" definition for `foo` and attach this metadata to it.
SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET = "dagster/auto_created_stub_asset"


@whitelist_for_serdes
class AssetExecutionType(Enum):
    OBSERVATION = "OBSERVATION"
    UNEXECUTABLE = "UNEXECUTABLE"
    MATERIALIZATION = "MATERIALIZATION"


@experimental_param(param="owners")
@experimental_param(param="tags")
class AssetSpec(
    NamedTuple(
        "_AssetSpec",
        [
            ("key", PublicAttr[AssetKey]),
            ("deps", PublicAttr[Iterable["AssetDep"]]),
            ("description", PublicAttr[Optional[str]]),
            ("metadata", PublicAttr[Optional[Mapping[str, Any]]]),
            ("group_name", PublicAttr[Optional[str]]),
            ("skippable", PublicAttr[bool]),
            ("code_version", PublicAttr[Optional[str]]),
            ("freshness_policy", PublicAttr[Optional[FreshnessPolicy]]),
            ("auto_materialize_policy", PublicAttr[Optional[AutoMaterializePolicy]]),
            ("owners", PublicAttr[Optional[Sequence[str]]]),
            ("tags", PublicAttr[Optional[Mapping[str, str]]]),
        ],
    )
):
    """Specifies the core attributes of an asset. This object is attached to the decorated
    function that defines how it materialized.

    Attributes:
        key (AssetKey): The unique identifier for this asset.
        deps (Optional[AbstractSet[AssetKey]]): The asset keys for the upstream assets that
            materializing this asset depends on.
        description (Optional[str]): Human-readable description of this asset.
        metadata (Optional[Dict[str, Any]]): A dict of static metadata for this asset.
            For example, users can provide information about the database table this
            asset corresponds to.
        skippable (bool): Whether this asset can be omitted during materialization, causing downstream
            dependencies to skip.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If
            not provided, the name "default" is used.
        code_version (Optional[str]): The version of the code for this specific asset,
            overriding the code version of the materialization function
        freshness_policy (Optional[FreshnessPolicy]): (Deprecated) A policy which indicates how up
            to date this asset is intended to be.
        auto_materialize_policy (Optional[AutoMaterializePolicy]): AutoMaterializePolicy to apply to
            the specified asset.
        backfill_policy (Optional[BackfillPolicy]): BackfillPolicy to apply to the specified asset.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the asset. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
    """

    def __new__(
        cls,
        key: CoercibleToAssetKey,
        *,
        deps: Optional[Iterable["CoercibleToAssetDep"]] = None,
        description: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        skippable: bool = False,
        group_name: Optional[str] = None,
        code_version: Optional[str] = None,
        freshness_policy: Optional[FreshnessPolicy] = None,
        auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
        owners: Optional[Sequence[str]] = None,
        tags: Optional[Mapping[str, str]] = None,
    ):
        from dagster._core.definitions.asset_dep import coerce_to_deps_and_check_duplicates

        key = AssetKey.from_coercible(key)
        asset_deps = coerce_to_deps_and_check_duplicates(deps, key)

        return super().__new__(
            cls,
            key=key,
            deps=asset_deps,
            description=check.opt_str_param(description, "description"),
            metadata=check.opt_mapping_param(metadata, "metadata", key_type=str),
            skippable=check.bool_param(skippable, "skippable"),
            group_name=check.opt_str_param(group_name, "group_name"),
            code_version=check.opt_str_param(code_version, "code_version"),
            freshness_policy=check.opt_inst_param(
                freshness_policy,
                "freshness_policy",
                FreshnessPolicy,
            ),
            auto_materialize_policy=check.opt_inst_param(
                auto_materialize_policy,
                "auto_materialize_policy",
                AutoMaterializePolicy,
            ),
            owners=check.opt_sequence_param(owners, "owners", of_type=str),
            tags=validate_tags_strict(tags),
        )



import datetime
from contextlib import contextmanager
from unittest import mock

import packaging.version
import pendulum
from typing_extensions import TypeAlias

_IS_PENDULUM_2 = (
    hasattr(pendulum, "__version__")
    and getattr(packaging.version.parse(getattr(pendulum, "__version__")), "major") == 2
)

_IS_PENDULUM_3 = (
    hasattr(pendulum, "__version__")
    and getattr(packaging.version.parse(getattr(pendulum, "__version__")), "major") == 3
)

# pendulum 1 has no __version__ property
_IS_PENDULUM_1 = not _IS_PENDULUM_2 and not _IS_PENDULUM_3

POST_TRANSITION = pendulum.tz.POST_TRANSITION if _IS_PENDULUM_3 else pendulum.POST_TRANSITION
PRE_TRANSITION = pendulum.tz.PRE_TRANSITION if _IS_PENDULUM_3 else pendulum.PRE_TRANSITION
TRANSITION_ERROR = pendulum.tz.TRANSITION_ERROR if _IS_PENDULUM_3 else pendulum.TRANSITION_ERROR


def pendulum_create_timezone(tz_name: str):
    if _IS_PENDULUM_3:
        from pendulum.tz.timezone import Timezone

        return Timezone(tz_name)
    else:
        return pendulum.tz.timezone(tz_name)


@contextmanager
def mock_pendulum_timezone(override_timezone):
    if _IS_PENDULUM_1:
        with pendulum.tz.LocalTimezone.test(pendulum.Timezone.load(override_timezone)):
            yield
    else:
        with pendulum.tz.test_local_timezone(pendulum.tz.timezone(override_timezone)):
            yield


def create_pendulum_time(year, month, day, *args, **kwargs):
    if "tz" in kwargs and "dst_rule" in kwargs and _IS_PENDULUM_1:
        tz_name_or_info = kwargs.pop("tz")
        tz = (
            pendulum.timezone(tz_name_or_info)
            if isinstance(tz_name_or_info, str)
            else tz_name_or_info
        )
        dst_rule = kwargs.pop("dst_rule")

        return pendulum.instance(
            tz.convert(
                datetime.datetime(
                    year,
                    month,
                    day,
                    *args,
                    **kwargs,
                ),
                dst_rule=dst_rule,
            )
        )

    if "dst_rule" in kwargs and _IS_PENDULUM_3:
        dst_rule = kwargs.pop("dst_rule")
        if dst_rule == PRE_TRANSITION:
            kwargs["fold"] = 0
        elif dst_rule == POST_TRANSITION:
            kwargs["fold"] = 1
        elif dst_rule == TRANSITION_ERROR:
            tz_name = kwargs.pop("tz")
            assert tz_name
            return pendulum.instance(
                pendulum_create_timezone(tz_name).convert(
                    datetime.datetime(
                        year,
                        month,
                        day,
                        *args,
                        **kwargs,
                    ),
                    raise_on_unknown_times=True,
                )
            )

    return (
        pendulum.create(year, month, day, *args, **kwargs)
        if _IS_PENDULUM_1
        else pendulum.datetime(year, month, day, *args, **kwargs)
    )


PendulumDateTime: TypeAlias = (
    pendulum.Pendulum if _IS_PENDULUM_1 else pendulum.DateTime  # type: ignore[attr-defined]
)

PendulumInterval: TypeAlias = (
    pendulum.Interval if _IS_PENDULUM_3 else pendulum.Period  # type: ignore[attr-defined]
)


@contextmanager
def pendulum_freeze_time(t):
    if _IS_PENDULUM_3:
        with mock.patch("pendulum.now", return_value=t):
            yield
    else:
        with pendulum.test(t) as frozen_time:
            yield frozen_time


# Workaround for issue with .in_tz() in pendulum:
# https://github.com/sdispater/pendulum/issues/535
def to_timezone(dt: PendulumDateTime, tz: str):
    timestamp = dt.timestamp()
    # handle negative timestamps, which are not natively supported on Windows
    if timestamp < 0:
        return pendulum.from_timestamp(0, tz=tz) + datetime.timedelta(seconds=timestamp)
    return pendulum.from_timestamp(dt.timestamp(), tz=tz)


def get_crontab_day_of_week(dt: PendulumDateTime) -> int:
    if _IS_PENDULUM_3:
        # In pendulum 3, 0-6 is Monday-Sunday (unlike crontab, where 0-6 is Sunday-Saturday)
        return (dt.day_of_week + 1) % 7
    else:
        return dt.day_of_week



from typing import Any, Mapping, Optional, cast

from dagster._annotations import experimental
from dagster._core.definitions.asset_selection import CoercibleToAssetSelection

from .asset_selection import AssetSelection
from .sensor_definition import DefaultSensorStatus, SensorDefinition, SensorType
from .utils import check_valid_name, normalize_tags


@experimental
class AutoMaterializeSensorDefinition(SensorDefinition):
    """Targets a set of assets and repeatedly evaluates all the AutoMaterializePolicys on all of
    those assets to determine which to request runs for.

    Args:
        name: The name of the sensor.
        asset_selection (Union[str, Sequence[str], Sequence[AssetKey], Sequence[Union[AssetsDefinition, SourceAsset]], AssetSelection]):
            The assets to evaluate AutoMaterializePolicys of and request runs for.
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        minimum_interval_seconds (Optional[int]): The frequency at which to try to evaluate the
            sensor. The actual interval will be longer if the sensor evaluation takes longer than
            the provided interval.
        description (Optional[str]): A human-readable description of the sensor.
    """

    def __init__(
        self,
        name: str,
        *,
        asset_selection: CoercibleToAssetSelection,
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
    ):
        self._run_tags = normalize_tags(run_tags).tags

        def evaluation_fn(context):
            raise NotImplementedError(
                "Automation policy sensors cannot be evaluated like regular user-space sensors."
            )

        super(AutoMaterializeSensorDefinition, self).__init__(
            name=check_valid_name(name),
            job_name=None,
            evaluation_fn=evaluation_fn,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            job=None,
            jobs=None,
            default_status=default_status,
            required_resource_keys=None,
            asset_selection=asset_selection,
        )

    @property
    def run_tags(self) -> Mapping[str, str]:
        return self._run_tags

    @property
    def asset_selection(self) -> AssetSelection:
        return cast(AssetSelection, super().asset_selection)

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.AUTO_MATERIALIZE



from typing import TYPE_CHECKING, Iterable, Mapping, Optional, Sequence, Set

import dagster._check as check
from dagster._core.definitions.events import AssetKey
from dagster._core.events.log import EventLogEntry
from dagster._core.instance import DagsterInstance

if TYPE_CHECKING:
    from dagster._core.storage.event_log.base import AssetRecord


class BatchAssetRecordLoader:
    """A batch loader that fetches asset records.  This loader is expected to be
    instantiated with a set of asset keys.
    """

    def __init__(self, instance: DagsterInstance, asset_keys: Iterable[AssetKey]):
        self._instance = instance
        self._unfetched_asset_keys: Set[AssetKey] = set(asset_keys)
        self._asset_records: Mapping[AssetKey, Optional["AssetRecord"]] = {}

    def add_asset_keys(self, asset_keys: Iterable[AssetKey]):
        unfetched_asset_keys = set(asset_keys).difference(self._asset_records)
        self._unfetched_asset_keys = self._unfetched_asset_keys.union(unfetched_asset_keys)

    def get_asset_record(self, asset_key: AssetKey) -> Optional["AssetRecord"]:
        if asset_key not in self._asset_records and asset_key not in self._unfetched_asset_keys:
            check.failed(
                f"Asset key {asset_key} not recognized for this loader. Expected one of:"
                f" {self._unfetched_asset_keys.union(self._asset_records.keys())}"
            )

        if asset_key in self._unfetched_asset_keys:
            self.fetch()

        return self._asset_records.get(asset_key)

    def clear_cache(self):
        """For use in tests."""
        self._unfetched_asset_keys = self._unfetched_asset_keys.union(self._asset_records.keys())
        self._asset_records = {}

    def has_cached_asset_record(self, asset_key: AssetKey):
        return asset_key in self._asset_records

    def get_asset_records(self, asset_keys: Sequence[AssetKey]) -> Sequence["AssetRecord"]:
        records = [self.get_asset_record(asset_key) for asset_key in asset_keys]
        return [record for record in records if record]

    def get_latest_materialization_for_asset_key(
        self, asset_key: AssetKey
    ) -> Optional[EventLogEntry]:
        asset_record = self.get_asset_record(asset_key)
        if not asset_record:
            return None

        return asset_record.asset_entry.last_materialization

    def get_latest_observation_for_asset_key(self, asset_key: AssetKey) -> Optional[EventLogEntry]:
        check.invariant(
            self._instance.event_log_storage.asset_records_have_last_observation,
            "Event log storage must support fetching the last observation from asset records",
        )

        asset_record = self.get_asset_record(asset_key)
        if not asset_record:
            return None

        return asset_record.asset_entry.last_observation

    def fetch(self) -> None:
        if not self._unfetched_asset_keys:
            return

        new_records = {
            record.asset_entry.asset_key: record
            for record in self._instance.get_asset_records(list(self._unfetched_asset_keys))
        }

        self._asset_records = {
            **self._asset_records,
            **{asset_key: new_records.get(asset_key) for asset_key in self._unfetched_asset_keys},
        }
        self._unfetched_asset_keys = set()



"""add run status index.

Revision ID: 3e71cf573ba6
Revises: 6d366a41b4be
Create Date: 2021-02-23 16:00:45.689578

"""

from alembic import op
from sqlalchemy import inspect

# revision identifiers, used by Alembic.
revision = "3e71cf573ba6"
down_revision = "6d366a41b4be"
branch_labels = None
depends_on = None


def upgrade():
    inspector = inspect(op.get_bind())
    has_tables = inspector.get_table_names()
    if "runs" in has_tables:
        indices = [x.get("name") for x in inspector.get_indexes("runs")]
        if "idx_run_status" not in indices:
            op.create_index("idx_run_status", "runs", ["status"], unique=False)


def downgrade():
    inspector = inspect(op.get_bind())
    has_tables = inspector.get_table_names()
    if "runs" in has_tables:
        indices = [x.get("name") for x in inspector.get_indexes("runs")]
        if "idx_run_status" in indices:
            op.drop_index("idx_run_status", "runs")



from abc import ABC, abstractproperty
from collections import defaultdict
from enum import Enum
from typing import (
    AbstractSet,
    Dict,
    FrozenSet,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataMapping, MetadataValue
from dagster._serdes.serdes import (
    _WHITELIST_MAP,
    NamedTupleSerializer,
    PackableValue,
    PydanticModelSerializer,
    UnpackContext,
    UnpackedValue,
    WhitelistMap,
    deserialize_value,
    whitelist_for_serdes,
)
from dagster._utils.security import non_secure_md5_hash_str

from .declarative_scheduling.serialized_objects import (
    AssetConditionEvaluation,
    AssetConditionEvaluationWithRunIds,
    AssetConditionSnapshot,
    AssetSubsetWithMetadata,
)
from .partition import PartitionsDefinition, SerializedPartitionsSubset


@whitelist_for_serdes
class AutoMaterializeDecisionType(Enum):
    """Represents the set of results of the auto-materialize logic.

    MATERIALIZE: The asset should be materialized by a run kicked off on this tick
    SKIP: The asset should not be materialized by a run kicked off on this tick, because future
        ticks are expected to materialize it.
    DISCARD: The asset should not be materialized by a run kicked off on this tick, but future
        ticks are not expected to materialize it.
    """

    MATERIALIZE = "MATERIALIZE"
    SKIP = "SKIP"
    DISCARD = "DISCARD"


@whitelist_for_serdes
class AutoMaterializeRuleSnapshot(NamedTuple):
    """A serializable snapshot of an AutoMaterializeRule for historical evaluations."""

    class_name: str
    description: str
    decision_type: AutoMaterializeDecisionType


class AutoMaterializeRuleEvaluationData(ABC):
    @abstractproperty
    def metadata(self) -> MetadataMapping:
        raise NotImplementedError()

    @property
    def frozen_metadata(self) -> FrozenSet[Tuple[str, MetadataValue]]:
        return frozenset(self.metadata.items())


@whitelist_for_serdes
class TextRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple("_TextRuleEvaluationData", [("text", str)]),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {"text": MetadataValue.text(self.text)}


@whitelist_for_serdes
class ParentUpdatedRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple(
        "_ParentUpdatedRuleEvaluationData",
        [
            ("updated_asset_keys", FrozenSet[AssetKey]),
            ("will_update_asset_keys", FrozenSet[AssetKey]),
        ],
    ),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {
            **{
                f"updated_parent_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.updated_asset_keys))
            },
            **{
                f"will_update_parent_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.will_update_asset_keys))
            },
        }


@whitelist_for_serdes
class WaitingOnAssetsRuleEvaluationData(
    AutoMaterializeRuleEvaluationData,
    NamedTuple(
        "_WaitingOnParentRuleEvaluationData",
        [("waiting_on_asset_keys", FrozenSet[AssetKey])],
    ),
):
    @property
    def metadata(self) -> MetadataMapping:
        return {
            **{
                f"waiting_on_ancestor_{i+1}": MetadataValue.asset(k)
                for i, k in enumerate(sorted(self.waiting_on_asset_keys))
            },
        }


RuleEvaluations = Tuple[AssetSubset, Sequence["AssetSubsetWithMetadata"], PackableValue]


@whitelist_for_serdes
class AutoMaterializeRuleEvaluation(NamedTuple):
    rule_snapshot: AutoMaterializeRuleSnapshot
    evaluation_data: Optional[AutoMaterializeRuleEvaluationData]


# BACKCOMPAT GRAVEYARD


def deserialize_auto_materialize_asset_evaluation_to_asset_condition_evaluation_with_run_ids(
    serialized_evaluation: str, partitions_def: Optional[PartitionsDefinition]
) -> "AssetConditionEvaluationWithRunIds":
    """Provides a backcompat layer to allow deserializing old AutoMaterializeAssetEvaluation
    objects into the new AssetConditionEvaluationWithRunIds objects.
    """
    from .declarative_scheduling.serialized_objects import (
        AssetConditionEvaluationWithRunIds,
    )

    class BackcompatDeserializer(BackcompatAutoMaterializeAssetEvaluationSerializer):
        @property
        def partitions_def(self) -> Optional[PartitionsDefinition]:
            return partitions_def

    # create a new WhitelistMap that can deserialize SerializedPartitionSubset objects stored
    # on the old cursor format
    whitelist_map = WhitelistMap(
        object_serializers=_WHITELIST_MAP.object_serializers,
        object_deserializers={
            **_WHITELIST_MAP.object_deserializers,
            "AutoMaterializeAssetEvaluation": BackcompatDeserializer(
                klass=AssetConditionEvaluationWithRunIds
            ),
        },
        enum_serializers=_WHITELIST_MAP.enum_serializers,
    )

    return deserialize_value(
        serialized_evaluation, AssetConditionEvaluationWithRunIds, whitelist_map=whitelist_map
    )


def deserialize_serialized_partitions_subset_to_asset_subset(
    serialized: SerializedPartitionsSubset,
    asset_key: AssetKey,
    partitions_def: Optional[PartitionsDefinition],
) -> AssetSubset:
    if partitions_def is None or not serialized.can_deserialize(partitions_def):
        # partitions def has changed since storage time
        return AssetSubset.empty(asset_key, partitions_def)

    return AssetSubset(asset_key=asset_key, value=serialized.deserialize(partitions_def))


class BackcompatAutoMaterializeAssetEvaluationSerializer(PydanticModelSerializer):
    """This handles backcompat for the old AutoMaterializeAssetEvaluation objects, turning them into
    AssetConditionEvaluationWithRunIds objects.
    """

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        """This property gets overridden by subclasses at runtime, once the partitions_def for the
        specific record we're deserializing is known.
        """
        raise NotImplementedError()

    def deserialize_serialized_partitions_subset_or_none(
        self,
        asset_key: AssetKey,
        serialized: Union[None, SerializedPartitionsSubset],
    ) -> AssetSubset:
        if serialized is None:
            # Confusingly, we used `None` to indicate "all of an unpartitioned asset" in the old
            # serialization scheme
            return AssetSubset(asset_key=asset_key, value=True)
        return deserialize_serialized_partitions_subset_to_asset_subset(
            serialized, asset_key, self.partitions_def
        )

    def _asset_condition_snapshot_from_rule_snapshot(
        self, rule_snapshot: AutoMaterializeRuleSnapshot
    ) -> "AssetConditionSnapshot":
        from .declarative_scheduling.legacy.rule_condition import RuleCondition
        from .declarative_scheduling.serialized_objects import AssetConditionSnapshot

        unique_id_parts = [rule_snapshot.class_name, rule_snapshot.description]
        unique_id = non_secure_md5_hash_str("".join(unique_id_parts).encode())

        return AssetConditionSnapshot(
            class_name=RuleCondition.__name__,
            description=rule_snapshot.description,
            unique_id=unique_id,
        )

    def _get_child_rule_evaluation(
        self,
        asset_key: AssetKey,
        partition_subsets_by_condition: Sequence[
            Tuple["AutoMaterializeRuleEvaluation", Optional[SerializedPartitionsSubset]]
        ],
        is_partitioned: bool,
        rule_snapshot: AutoMaterializeRuleSnapshot,
    ) -> "AssetConditionEvaluation":
        from .declarative_scheduling.serialized_objects import (
            HistoricalAllPartitionsSubsetSentinel,
        )

        condition_snapshot = self._asset_condition_snapshot_from_rule_snapshot(rule_snapshot)

        subsets_with_metadata = [
            AssetSubsetWithMetadata(
                subset=self.deserialize_serialized_partitions_subset_or_none(asset_key, serialized),
                metadata=rule_evaluation.evaluation_data.metadata,
            )
            for rule_evaluation, serialized in partition_subsets_by_condition
            if rule_evaluation.evaluation_data
        ]

        true_subset = AssetSubset.empty(asset_key, self.partitions_def)
        for _, serialized in partition_subsets_by_condition:
            true_subset |= self.deserialize_serialized_partitions_subset_or_none(
                asset_key, serialized
            )

        return AssetConditionEvaluation(
            condition_snapshot=condition_snapshot,
            true_subset=true_subset,
            candidate_subset=HistoricalAllPartitionsSubsetSentinel()
            if is_partitioned
            else AssetSubset.all(asset_key, None),
            start_timestamp=None,
            end_timestamp=None,
            subsets_with_metadata=subsets_with_metadata,
            child_evaluations=[],
        )

    def _get_child_decision_type_evaluation(
        self,
        asset_key: AssetKey,
        partition_subsets_by_condition: Sequence[
            Tuple["AutoMaterializeRuleEvaluation", Optional[SerializedPartitionsSubset]]
        ],
        rule_snapshots: Sequence[AutoMaterializeRuleSnapshot],
        is_partitioned: bool,
        decision_type: AutoMaterializeDecisionType,
    ) -> Optional["AssetConditionEvaluation"]:
        from .declarative_scheduling.operators.boolean_operators import (
            NotAssetCondition,
            OrAssetCondition,
        )
        from .declarative_scheduling.serialized_objects import (
            HistoricalAllPartitionsSubsetSentinel,
        )

        partition_subsets_by_condition_by_rule_snapshot = defaultdict(list)
        for elt in partition_subsets_by_condition:
            partition_subsets_by_condition_by_rule_snapshot[elt[0].rule_snapshot].append(elt)

        child_evaluations = [
            self._get_child_rule_evaluation(
                asset_key,
                partition_subsets_by_condition_by_rule_snapshot[rule_snapshot],
                is_partitioned,
                rule_snapshot,
            )
            for rule_snapshot in (
                set(rule_snapshots) | set(partition_subsets_by_condition_by_rule_snapshot.keys())
            )
            if rule_snapshot.decision_type == decision_type
        ]

        if decision_type == AutoMaterializeDecisionType.DISCARD:
            # for the discard type, we don't have an OrAssetCondition
            if len(child_evaluations) != 1:
                return None
            evaluation = child_evaluations[0]
        else:
            unique_id_parts = [
                OrAssetCondition.__name__,
                *[e.condition_snapshot.unique_id for e in child_evaluations],
            ]
            unique_id = non_secure_md5_hash_str("".join(unique_id_parts).encode())
            decision_type_snapshot = AssetConditionSnapshot(
                class_name=OrAssetCondition.__name__, description="Any of", unique_id=unique_id
            )
            true_subset = AssetSubset.empty(asset_key, self.partitions_def)
            for e in child_evaluations:
                true_subset |= e.true_subset
            evaluation = AssetConditionEvaluation(
                condition_snapshot=decision_type_snapshot,
                true_subset=true_subset,
                candidate_subset=HistoricalAllPartitionsSubsetSentinel()
                if is_partitioned
                else AssetSubset.all(asset_key, None),
                subsets_with_metadata=[],
                child_evaluations=child_evaluations,
                start_timestamp=None,
                end_timestamp=None,
            )

        if decision_type == AutoMaterializeDecisionType.MATERIALIZE:
            return evaluation

        # non-materialize conditions are inverted
        unique_id_parts = [
            NotAssetCondition.__name__,
            evaluation.condition_snapshot.unique_id,
        ]
        unique_id = non_secure_md5_hash_str("".join(unique_id_parts).encode())

        if is_partitioned:
            # In reality, we'd like to invert the inner true_subset here, but this is an
            # expensive operation, and error-prone as the set of all partitions may have changed
            # since the evaluation was stored. Instead, we just use an empty subset.
            true_subset = AssetSubset.empty(asset_key, self.partitions_def)
        else:
            true_subset = evaluation.true_subset.copy(
                update={"value": not evaluation.true_subset.bool_value}
            )
        return AssetConditionEvaluation(
            condition_snapshot=AssetConditionSnapshot(
                class_name=NotAssetCondition.__name__, description="Not", unique_id=unique_id
            ),
            true_subset=true_subset,
            candidate_subset=HistoricalAllPartitionsSubsetSentinel()
            if is_partitioned
            else AssetSubset.all(asset_key, None),
            subsets_with_metadata=[],
            child_evaluations=[evaluation],
            start_timestamp=None,
            end_timestamp=None,
        )

    def unpack(
        self,
        unpacked_dict: Dict[str, UnpackedValue],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> "AssetConditionEvaluationWithRunIds":
        from .declarative_scheduling.operators.boolean_operators import AndAssetCondition
        from .declarative_scheduling.serialized_objects import (
            HistoricalAllPartitionsSubsetSentinel,
        )

        asset_key = cast(AssetKey, unpacked_dict.get("asset_key"))
        partition_subsets_by_condition = cast(
            Sequence[Tuple[AutoMaterializeRuleEvaluation, Optional[SerializedPartitionsSubset]]],
            unpacked_dict.get("partition_subsets_by_condition"),
        )
        rule_snapshots = (
            cast(Sequence[AutoMaterializeRuleSnapshot], unpacked_dict.get("rule_snapshots", []))
            or []
        )
        is_partitioned = (
            any(tup[1] is not None for tup in partition_subsets_by_condition)
            if partition_subsets_by_condition
            # if we don't have any partition_subsets_by_condition to look at, we can't tell if this
            # asset was partitioned at the time that the evaluation was stored, so instead we assume
            # that its current partition status is the same as its partition status at storage time.
            else self.partitions_def is not None
        )

        # get the sub-evaluations for each decision type
        materialize_evaluation = check.not_none(
            self._get_child_decision_type_evaluation(
                asset_key,
                partition_subsets_by_condition,
                rule_snapshots,
                is_partitioned,
                AutoMaterializeDecisionType.MATERIALIZE,
            )
        )
        not_skip_evaluation = self._get_child_decision_type_evaluation(
            asset_key,
            partition_subsets_by_condition,
            rule_snapshots,
            is_partitioned,
            AutoMaterializeDecisionType.SKIP,
        )
        not_discard_evaluation = self._get_child_decision_type_evaluation(
            asset_key,
            partition_subsets_by_condition,
            rule_snapshots,
            is_partitioned,
            AutoMaterializeDecisionType.DISCARD,
        )

        # filter out any None evaluations (should realistically only happen for discard)
        child_evaluations = list(
            filter(None, [materialize_evaluation, not_skip_evaluation, not_discard_evaluation])
        )

        # the top level condition is the AND of all the sub-conditions
        unique_id_parts = [
            AndAssetCondition.__name__,
            *[e.condition_snapshot.unique_id for e in child_evaluations],
        ]
        unique_id = non_secure_md5_hash_str("".join(unique_id_parts).encode())
        condition_snapshot = AssetConditionSnapshot(
            class_name=AndAssetCondition.__name__, description="All of", unique_id=unique_id
        )

        # all AssetSubsets generated here are created using the current partitions_def, so they
        # will be valid
        true_subset = materialize_evaluation.true_subset.as_valid(self.partitions_def)
        if not_skip_evaluation:
            true_subset -= not_skip_evaluation.child_evaluations[0].true_subset
        if not_discard_evaluation:
            true_subset -= not_discard_evaluation.child_evaluations[0].true_subset

        return AssetConditionEvaluation(
            condition_snapshot=condition_snapshot,
            true_subset=true_subset,
            candidate_subset=HistoricalAllPartitionsSubsetSentinel()
            if is_partitioned
            else AssetSubset.all(asset_key, None),
            subsets_with_metadata=[],
            child_evaluations=child_evaluations,
            start_timestamp=None,
            end_timestamp=None,
        ).with_run_ids(cast(AbstractSet[str], unpacked_dict.get("run_ids", set())))


class BackcompatAutoMaterializeConditionSerializer(NamedTupleSerializer):
    """This handles backcompat for the old AutoMaterializeCondition objects, turning them into the
    proper AutoMaterializeRuleEvaluation objects. This is necessary because old
    AutoMaterializeAssetEvaluation objects will have serialized AutoMaterializeCondition objects,
    and we need to be able to deserialize them.

    In theory, as these serialized objects happen to be purged periodically, we can remove this
    backcompat logic at some point in the future.
    """

    def unpack(
        self,
        unpacked_dict: Dict[str, UnpackedValue],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> AutoMaterializeRuleEvaluation:
        from .auto_materialize_rule import AutoMaterializeRule
        from .auto_materialize_rule_impls import DiscardOnMaxMaterializationsExceededRule

        if self.klass in (
            FreshnessAutoMaterializeCondition,
            DownstreamFreshnessAutoMaterializeCondition,
        ):
            return AutoMaterializeRuleEvaluation(
                rule_snapshot=AutoMaterializeRule.materialize_on_required_for_freshness().to_snapshot(),
                evaluation_data=None,
            )
        elif self.klass == MissingAutoMaterializeCondition:
            return AutoMaterializeRuleEvaluation(
                rule_snapshot=AutoMaterializeRule.materialize_on_missing().to_snapshot(),
                evaluation_data=None,
            )
        elif self.klass == ParentMaterializedAutoMaterializeCondition:
            updated_asset_keys = unpacked_dict.get("updated_asset_keys")
            if isinstance(updated_asset_keys, set):
                updated_asset_keys = cast(FrozenSet[AssetKey], frozenset(updated_asset_keys))
            else:
                updated_asset_keys = frozenset()
            will_update_asset_keys = unpacked_dict.get("will_update_asset_keys")
            if isinstance(will_update_asset_keys, set):
                will_update_asset_keys = cast(
                    FrozenSet[AssetKey], frozenset(will_update_asset_keys)
                )
            else:
                will_update_asset_keys = frozenset()
            return AutoMaterializeRuleEvaluation(
                rule_snapshot=AutoMaterializeRule.materialize_on_parent_updated().to_snapshot(),
                evaluation_data=ParentUpdatedRuleEvaluationData(
                    updated_asset_keys=updated_asset_keys,
                    will_update_asset_keys=will_update_asset_keys,
                ),
            )
        elif self.klass == ParentOutdatedAutoMaterializeCondition:
            waiting_on_asset_keys = unpacked_dict.get("waiting_on_asset_keys")
            if isinstance(waiting_on_asset_keys, set):
                waiting_on_asset_keys = cast(FrozenSet[AssetKey], frozenset(waiting_on_asset_keys))
            else:
                waiting_on_asset_keys = frozenset()
            return AutoMaterializeRuleEvaluation(
                rule_snapshot=AutoMaterializeRule.skip_on_parent_outdated().to_snapshot(),
                evaluation_data=WaitingOnAssetsRuleEvaluationData(
                    waiting_on_asset_keys=waiting_on_asset_keys
                ),
            )
        elif self.klass == MaxMaterializationsExceededAutoMaterializeCondition:
            return AutoMaterializeRuleEvaluation(
                rule_snapshot=DiscardOnMaxMaterializationsExceededRule(limit=1).to_snapshot(),
                evaluation_data=None,
            )
        check.failed(f"Unexpected class {self.klass}")


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class FreshnessAutoMaterializeCondition(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class DownstreamFreshnessAutoMaterializeCondition(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class ParentMaterializedAutoMaterializeCondition(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class MissingAutoMaterializeCondition(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class ParentOutdatedAutoMaterializeCondition(NamedTuple): ...


@whitelist_for_serdes(serializer=BackcompatAutoMaterializeConditionSerializer)
class MaxMaterializationsExceededAutoMaterializeCondition(NamedTuple): ...



import os
import random
import re
import string
import uuid
import warnings
from collections import OrderedDict
from concurrent.futures import Future, ThreadPoolExecutor
from contextvars import copy_context
from typing import (
    AbstractSet,
    Any,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    cast,
)
from weakref import WeakSet

import toposort as toposort_
from typing_extensions import Final

import dagster._check as check
from dagster._utils import library_version_from_core_version, parse_package_version

BACKFILL_TAG_LENGTH = 8

PYTHON_LOGGING_LEVELS_MAPPING: Final[Mapping[str, int]] = OrderedDict(
    {"CRITICAL": 50, "ERROR": 40, "WARNING": 30, "INFO": 20, "DEBUG": 10}
)

PYTHON_LOGGING_LEVELS_ALIASES: Final[Mapping[str, str]] = OrderedDict(
    {"FATAL": "CRITICAL", "WARN": "WARNING"}
)

PYTHON_LOGGING_LEVELS_NAMES = frozenset(
    [
        level_name.lower()
        for level_name in sorted(
            list(PYTHON_LOGGING_LEVELS_MAPPING.keys()) + list(PYTHON_LOGGING_LEVELS_ALIASES.keys())
        )
    ]
)

T = TypeVar("T", bound=Any)


def coerce_valid_log_level(log_level: Union[str, int]) -> int:
    """Convert a log level into an integer for consumption by the low-level Python logging API."""
    if isinstance(log_level, int):
        return log_level
    str_log_level = check.str_param(log_level, "log_level")
    check.invariant(
        str_log_level.lower() in PYTHON_LOGGING_LEVELS_NAMES,
        "Bad value for log level {level}: permissible values are {levels}.".format(
            level=str_log_level,
            levels=", ".join(
                [f"'{level_name.upper()}'" for level_name in PYTHON_LOGGING_LEVELS_NAMES]
            ),
        ),
    )
    str_log_level = PYTHON_LOGGING_LEVELS_ALIASES.get(log_level.upper(), log_level.upper())
    return PYTHON_LOGGING_LEVELS_MAPPING[str_log_level]


def toposort(data: Mapping[T, AbstractSet[T]]) -> Sequence[Sequence[T]]:
    # Workaround a bug in older versions of toposort that choke on frozenset
    data = {k: set(v) if isinstance(v, frozenset) else v for k, v in data.items()}
    return [sorted(list(level)) for level in toposort_.toposort(data)]


def toposort_flatten(data: Mapping[T, AbstractSet[T]]) -> Sequence[T]:
    return [item for level in toposort(data) for item in level]


def make_new_run_id() -> str:
    return str(uuid.uuid4())


def make_new_backfill_id() -> str:
    return "".join(random.choice(string.ascii_lowercase) for x in range(BACKFILL_TAG_LENGTH))


def str_format_list(items: Iterable[object]) -> str:
    return "[{items}]".format(items=", ".join([f"'{item}'" for item in items]))


def str_format_set(items: Iterable[object]) -> str:
    return "[{items}]".format(items=", ".join([f"'{item}'" for item in items]))


def check_dagster_package_version(library_name: str, library_version: str) -> None:
    # This import must be internal in order for this function to be testable
    from dagster.version import __version__

    parsed_lib_version = parse_package_version(library_version)
    if parsed_lib_version.release[0] >= 1:
        if library_version != __version__:
            message = (
                f"Found version mismatch between `dagster` ({__version__})"
                f"and `{library_name}` ({library_version})"
            )
            warnings.warn(message)
    else:
        target_version = library_version_from_core_version(__version__)
        if library_version != target_version:
            message = (
                f"Found version mismatch between `dagster` ({__version__}) "
                f"expected library version ({target_version}) "
                f"and `{library_name}` ({library_version})."
            )
            warnings.warn(message)


def get_env_var_name(env_var_str: str):
    if "=" in env_var_str:
        return env_var_str.split("=", maxsplit=1)[0]
    else:
        return env_var_str


def parse_env_var(env_var_str: str) -> Tuple[str, str]:
    if "=" in env_var_str:
        split = env_var_str.split("=", maxsplit=1)
        return (split[0], split[1])
    else:
        env_var_value = os.getenv(env_var_str)
        if env_var_value is None:
            raise Exception(f"Tried to load environment variable {env_var_str}, but it was not set")
        return (env_var_str, cast(str, env_var_value))


class RequestUtilizationMetrics(TypedDict):
    """A dict of utilization metrics for a threadpool executor. We use generic language in case we use this metrics in a scenario where we switch away from a threadpool executor at a later time."""

    max_concurrent_requests: Optional[int]
    num_running_requests: Optional[int]
    num_queued_requests: Optional[int]


class FuturesAwareThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        thread_name_prefix: str = "",
        initializer: Any = None,
        initargs: Tuple[Any, ...] = (),
    ) -> None:
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
        # The default threadpool class doesn't track the futures it creates,
        # so if we want to be able to count the number of running futures, we need to do it ourselves.
        self._tracked_futures: WeakSet[Future] = WeakSet()

    def submit(self, fn, *args, **kwargs) -> Future:
        new_future = super().submit(fn, *args, **kwargs)
        self._tracked_futures.add(new_future)
        return new_future

    @property
    def max_workers(self) -> int:
        return self._max_workers

    @property
    def weak_tracked_futures_count(self) -> int:
        return len(self._tracked_futures)

    @property
    def num_running_futures(self) -> int:
        return sum(1 for f in self._tracked_futures if not f.done()) - self.num_queued_futures

    @property
    def num_queued_futures(self) -> int:
        return self._work_queue.qsize()

    def get_current_utilization_metrics(self) -> RequestUtilizationMetrics:
        return {
            "max_concurrent_requests": self.max_workers,
            "num_running_requests": self.num_running_futures,
            "num_queued_requests": self.num_queued_futures,
        }


class InheritContextThreadPoolExecutor(FuturesAwareThreadPoolExecutor):
    """A ThreadPoolExecutor that copies over contextvars at submit time."""

    def submit(self, fn, *args, **kwargs):
        ctx = copy_context()
        return super().submit(ctx.run, fn, *args, **kwargs)


def is_valid_email(email: str) -> bool:
    regex = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b"
    return bool(re.fullmatch(regex, email))



from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, Hashable, Optional

from pydantic import BaseModel, ConfigDict, PrivateAttr
from typing_extensions import Annotated, Self, TypeAlias, TypeVar

from .pydantic_compat_layer import USING_PYDANTIC_2

if USING_PYDANTIC_2:
    from pydantic import InstanceOf as InstanceOf  # type: ignore
else:
    # fallback to a no-op on pydantic 1 as there is no equivalent
    AnyType = TypeVar("AnyType")
    InstanceOf: TypeAlias = Annotated[AnyType, ...]


class DagsterModel(BaseModel):
    """Standardizes on Pydantic settings that are stricter than the default.
    - Frozen, to avoid complexity caused by mutation.
    - extra=forbid, to avoid bugs caused by accidentally constructing with the wrong arguments.
    - arbitrary_types_allowed, to allow non-model class params to be validated with isinstance.
    - Avoid pydantic reading a cached property class as part of the schema.
    """

    if not USING_PYDANTIC_2:
        # the setattr approach for cached_method works in pydantic 2 so only declare the PrivateAttr
        # in pydantic 1 as it has non trivial performance impact
        _cached_method_cache__internal__: Dict[Hashable, Any] = PrivateAttr(default_factory=dict)

    if TYPE_CHECKING:
        # without this, the type checker does not understand the constructor kwargs on subclasses
        def __init__(self, **data: Any) -> None: ...

    if USING_PYDANTIC_2:
        model_config = ConfigDict(  # type: ignore
            extra="forbid",
            frozen=True,
            arbitrary_types_allowed=True,
            ignored_types=(cached_property,),
        )
    else:

        class Config:
            extra = "forbid"
            frozen = True
            arbitrary_types_allowed = True
            keep_untouched = (cached_property,)

    def model_copy(self, *, update: Optional[Dict[str, Any]] = None) -> Self:
        if USING_PYDANTIC_2:
            return super().model_copy(update=update)  # type: ignore
        else:
            return super().copy(update=update)

    @classmethod
    def model_construct(cls, **kwargs: Any) -> Self:
        if USING_PYDANTIC_2:
            return super().model_construct(**kwargs)  # type: ignore
        else:
            return super().construct(**kwargs)



import logging
from typing import Optional

from dagster._utils.error import (
    ExceptionInfo,
    SerializableErrorInfo,
    serializable_error_info_from_exc_info,
)


class DaemonErrorCapture:
    @staticmethod
    def default_on_exception(
        exc_info: ExceptionInfo,
        logger: Optional[logging.Logger] = None,
        log_message: Optional[str] = None,
    ) -> SerializableErrorInfo:
        error_info = serializable_error_info_from_exc_info(exc_info)
        if logger and log_message:
            logger.exception(log_message)
        return error_info

    # global behavior for how to handle unexpected exceptions
    on_exception = default_on_exception
